/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *  Copyright (C) 2008  g新部 Hiroyuki Ikezoe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implwebkit_gtkd warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <glib/gi18n.h>

#include "kz-webkit-gtk.h"

#include <webkit/webkit.h>

#include "kz-embed.h"

typedef struct _KzWebKitGtkPrivate	KzWebKitGtkPrivate;
struct _KzWebKitGtkPrivate
{
	WebKitWebView *web_view;
	gint progress;
	gchar *location;
	gchar *title;
	gchar *link_message;
	gboolean lock;
	gboolean is_loading;
};
#define KZ_WEBKIT_GTK_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), KZ_TYPE_WEBKIT_GTK, KzWebKitGtkPrivate))

static GType           kz_type_webkit_gtk = 0;
static GtkScrolledWindowClass *kz_webkit_gtk_parent_class;

GtkWidget  *kz_webkit_gtk_new          (void);
static void kz_webkit_gtk_class_init   (KzWebKitGtkClass *klass);
static void kz_webkit_gtk_iface_init   (KzEmbedIFace     *iface);
static void kz_webkit_gtk_init         (KzWebKitGtk      *webkit_gtk);
static void kz_webkit_gtk_dispose      (GObject          *object);

static GObject *constructor  (GType type,
			      guint n_props,
			      GObjectConstructParam *props);

static void	    load_url               (KzEmbed      *kzembed,
					    const gchar  *url);
static void	    stop_load              (KzEmbed      *kzembed);
static void	    reload		   (KzEmbed      *kzembed,
					    KzEmbedReloadFlag flags);
static void	    go_back                (KzEmbed      *kzembed);
static void	    go_forward             (KzEmbed      *kzembed);
static gboolean	    can_go_back            (KzEmbed      *kzembed);
static gboolean	    can_go_forward         (KzEmbed      *kzembed);
static gboolean     can_go_nav_link	   (KzEmbed      *kzembed,
					    KzEmbedNavLink link);
static KzNavi      *get_nth_nav_link	   (KzEmbed      *kzembed,
					    KzEmbedNavLink link,
					    guint         n);
static GList       *get_nav_links	   (KzEmbed      *kzembed,
					    KzEmbedNavLink link);
static gboolean     can_cut_selection      (KzEmbed      *kzembed);
static gboolean     can_copy_selection     (KzEmbed      *kzembed);
static gboolean     can_paste              (KzEmbed      *kzembed);
static void	    cut_selection          (KzEmbed      *kzembed);
static void	    copy_selection         (KzEmbed      *kzembed);
static void	    paste                  (KzEmbed      *kzembed);
static void	    select_all             (KzEmbed      *kzembed);
static gdouble	    get_progress	   (KzEmbed      *kzembed);
static const gchar *get_title              (KzEmbed      *kzembed);
static const gchar *get_location           (KzEmbed      *kzembed);
static gchar       *get_link_message       (KzEmbed      *kzembed);
static gboolean     selection_is_collapsed (KzEmbed      *kzembed);
static gboolean     is_loading             (KzEmbed      *kzembed);
static void         zoom_set               (KzEmbed      *kzembed, 
					    int           zoom, 
					    gboolean      reflow);
static int          zoom_get               (KzEmbed      *kzembed);
static gboolean     get_lock               (KzEmbed      *kzembed);
static void         set_lock               (KzEmbed      *kzembed,
					    gboolean      lock);
static gboolean	    get_allow_javascript   (KzEmbed      *kzembed);
static void	    set_allow_javascript   (KzEmbed      *kzembed,
    					    gboolean      allow);
static gboolean	    get_allow_images	   (KzEmbed      *kzembed);
static void	    set_allow_images	   (KzEmbed      *kzembed,
					    gboolean      allow);
static void	    get_history            (KzEmbed      *kzembed,
					    KzBookmarkFolder   *history);
static void	    set_history            (KzEmbed      *kzembed,
					    KzBookmarkFolder   *history);
static void         copy_page              (KzEmbed      *src_embed,
					    KzEmbed      *dest_embed);
static gboolean     shistory_get_pos       (KzEmbed      *kzembed,
					    gint         *pos,
					    gint         *count);
static void         shistory_get_nth       (KzEmbed      *kzembed, 
					    gint          nth,
					    gboolean      is_relative,
					    gchar       **uri,
					    gchar       **title);
static gboolean	    find		   (KzEmbed	 *kzembed,
					    const char   *keyword,
					    gboolean      backward);
static void	cb_title_changed   (WebKitWebView* web_view,
				    WebKitWebFrame* web_frame,
				    const gchar* title,
				    gpointer data);
static void	cb_progress_changed(WebKitWebView* web_view,
				    gint progress,
				    gpointer data);
static void	cb_load_committed  (WebKitWebView* web_view,
				    WebKitWebFrame* frame,
				    gpointer data);
static void	cb_load_started    (WebKitWebView* web_view,
				    WebKitWebFrame* frame,
				    gpointer data);
static void	cb_load_finished   (WebKitWebView* web_view,
				    WebKitWebFrame* frame,
				    gpointer data);
static void	cb_hover_link      (WebKitWebView* web_view,
				    const gchar *title,
				    const gchar *location,
				    gpointer data);
static void     cb_populate_popup  (WebKitWebView* web_view,
		                    GtkMenu *menu,
                                    gpointer data);
static gboolean cb_navigation_policy_decision_requested
                                   (WebKitWebView* web_view,
                                    WebKitWebFrame *frame,
                                    WebKitNetworkRequest *request,
                                    WebKitWebNavigationAction *action,
                                    WebKitWebPolicyDecision *decision,
                                    gpointer data);
static gboolean cb_button_press_event
                                   (GtkWidget *widget,
				    GdkEventButton *event,
                                    gpointer data);
static gboolean cb_button_release_event
                                   (GtkWidget *widget,
				    GdkEventButton *event,
                                    gpointer data);

static void set_default_preferences (KzWebKitGtk *webkit);

void
kz_webkit_gtk_register_type (GTypeModule *module)
{
	static const GTypeInfo kz_webkit_gtk_info =
	{
		sizeof (KzWebKitGtkClass),
		NULL,		/* base_init */
		NULL,		/* base_finalize */
		(GClassInitFunc) kz_webkit_gtk_class_init,
		NULL,		/* class_finalize */
		NULL,		/* class_data */
		sizeof (KzWebKitGtk),
		0,		/* n_preallocs */
		(GInstanceInitFunc) kz_webkit_gtk_init,
	};

	const GInterfaceInfo kz_embed_info =
	{
		(GInterfaceInitFunc) kz_webkit_gtk_iface_init,
		NULL,
		NULL
	};

	kz_type_webkit_gtk = g_type_module_register_type(module,
						 GTK_TYPE_SCROLLED_WINDOW,
						 "KzWebKitGtk",
						 &kz_webkit_gtk_info,
						 (GTypeFlags)0);

	g_type_module_add_interface(module,
				    KZ_TYPE_WEBKIT_GTK,
				    KZ_TYPE_EMBED,
				    &kz_embed_info);
}

GType
kz_webkit_gtk_get_type (void)
{
       return kz_type_webkit_gtk;
}

static void
kz_webkit_gtk_class_init (KzWebKitGtkClass *klass)
{
	GObjectClass *object_class;
	GtkWidgetClass *widget_class;

	kz_webkit_gtk_parent_class = GTK_SCROLLED_WINDOW_CLASS(g_type_class_peek_parent(klass));
	object_class = G_OBJECT_CLASS(klass);
	widget_class = GTK_WIDGET_CLASS(klass);

	object_class->dispose = kz_webkit_gtk_dispose;
	object_class->constructor = constructor;

	g_type_class_add_private(object_class, sizeof(KzWebKitGtkPrivate));
}

static void
kz_webkit_gtk_iface_init (KzEmbedIFace *iface)
{
	iface->load_url               = load_url;
	iface->view_source            = NULL;
	iface->view_current_page_source_in_new
				      = NULL;
	iface->is_loading             = is_loading;
	iface->get_title              = get_title;
	iface->get_location           = get_location;
	iface->get_link_message       = get_link_message;
	iface->get_progress           = get_progress;
	iface->can_cut_selection      = can_cut_selection;
	iface->can_copy_selection     = can_copy_selection;
	iface->can_paste              = can_paste;
	iface->cut_selection          = cut_selection;
	iface->copy_selection         = copy_selection;
	iface->paste                  = paste;
	iface->select_all             = select_all;
	iface->get_selection_string   = NULL;
	iface->find                   = find;
	iface->incremental_search     = NULL;
	iface->selection_is_collapsed = selection_is_collapsed;
	iface->get_links              = NULL;
	iface->copy_page              = copy_page;
	iface->shistory_get_pos       = shistory_get_pos;
	iface->shistory_get_nth       = shistory_get_nth;
	iface->reload                 = reload;
	iface->stop_load              = stop_load;
	iface->go_back                = go_back;
	iface->go_forward             = go_forward;
	iface->can_go_back            = can_go_back;
	iface->can_go_forward         = can_go_forward;
	iface->can_go_nav_link        = can_go_nav_link;
	iface->go_nav_link            = NULL;
	iface->append_nav_link        = NULL;
	iface->set_nav_link           = NULL;
	iface->set_nth_nav_link       = NULL;
	iface->get_nav_link           = NULL;
	iface->get_nth_nav_link       = get_nth_nav_link;
	iface->get_nav_links          = get_nav_links;
	iface->go_history_index       = NULL;
	iface->do_command             = NULL;
	iface->can_do_command         = NULL;
	iface->get_lock               = get_lock;
	iface->set_lock               = set_lock;
	iface->get_body_text          = NULL;
	iface->set_encoding           = NULL;
	iface->get_encoding           = NULL;
	iface->print                  = NULL;
	iface->print_preview          = NULL;
	iface->get_printer_list       = NULL;
	iface->create_thumbnail       = NULL;
	iface->save_with_content      = NULL;
	iface->set_text_into_textarea = NULL;
	iface->get_text_from_textarea = NULL;
	iface->zoom_set               = zoom_set;
	iface->zoom_get               = zoom_get;
	iface->set_text_size          = NULL;
	iface->get_text_size          = NULL;
	iface->get_html_with_contents = NULL;
	iface->set_history            = set_history;
	iface->get_history            = get_history;
	iface->get_last_modified      = NULL;
	iface->fine_scroll            = NULL;
	iface->page_up                = NULL;
	iface->page_down              = NULL; 
	iface->get_allow_javascript   = get_allow_javascript;
	iface->set_allow_javascript   = set_allow_javascript; 
	iface->get_allow_images       = get_allow_images;
	iface->set_allow_images       = set_allow_images;

	iface->link_message           = NULL;
	iface->js_status              = NULL;
	iface->location               = NULL;
	iface->title                  = NULL;
	iface->progress               = NULL;
	iface->net_start              = NULL;
	iface->net_stop               = NULL;
	iface->new_window             = NULL;
	iface->open_uri               = NULL;
	iface->size_to                = NULL;
	iface->dom_key_down           = NULL;
	iface->dom_key_press          = NULL;
	iface->dom_key_up             = NULL;
	iface->dom_mouse_down         = NULL;
	iface->dom_mouse_up           = NULL;
	iface->dom_mouse_click        = NULL;
	iface->dom_mouse_dbl_click    = NULL;
	iface->dom_mouse_over         = NULL;
	iface->dom_mouse_out          = NULL;
	iface->security_change        = NULL;
	iface->status_change          = NULL;
}

static void
setup_webkit_web_view (KzWebKitGtk *webkit)
{
	KzWebKitGtkPrivate *priv = KZ_WEBKIT_GTK_GET_PRIVATE(webkit);

	priv->web_view = WEBKIT_WEB_VIEW(webkit_web_view_new());
	gtk_container_add(GTK_CONTAINER(webkit), GTK_WIDGET(priv->web_view));

	g_signal_connect(priv->web_view, "title-changed",
			 G_CALLBACK(cb_title_changed), webkit);
	g_signal_connect(priv->web_view, "load-progress-changed",
			 G_CALLBACK(cb_progress_changed), webkit);
	g_signal_connect(priv->web_view, "load-committed",
			 G_CALLBACK(cb_load_committed), webkit);
	g_signal_connect(priv->web_view, "load-started",
			 G_CALLBACK(cb_load_started), webkit);
	g_signal_connect(priv->web_view, "load-finished",
			 G_CALLBACK(cb_load_finished), webkit);
	g_signal_connect(priv->web_view, "hovering-over-link",
			 G_CALLBACK(cb_hover_link), webkit);
	g_signal_connect(priv->web_view, "populate-popup",
			 G_CALLBACK(cb_populate_popup), webkit);
	g_signal_connect(priv->web_view, "navigation-policy-decision-requested",
			 G_CALLBACK(cb_navigation_policy_decision_requested), webkit);
	g_signal_connect(priv->web_view, "button-press-event",
			 G_CALLBACK(cb_button_press_event), webkit);
	g_signal_connect(priv->web_view, "button-release-event",
			 G_CALLBACK(cb_button_release_event), webkit);
	set_default_preferences(webkit);
}

static GObject*
constructor (GType                  type,
             guint                  n_props,
             GObjectConstructParam *props)
{
	GObject *object;
	GObjectClass *klass = G_OBJECT_CLASS(kz_webkit_gtk_parent_class);

	object = klass->constructor(type, n_props, props);
	setup_webkit_web_view(KZ_WEBKIT_GTK(object));

	return object;
}

static void
kz_webkit_gtk_init (KzWebKitGtk *webkit)
{
	KzWebKitGtkPrivate *priv = KZ_WEBKIT_GTK_GET_PRIVATE(webkit);

	priv->location = NULL;
	priv->title = NULL;
	priv->link_message = NULL;

	priv->progress = 0;

	priv->lock = FALSE;
	priv->is_loading = FALSE;
}

static void
kz_webkit_gtk_dispose (GObject *object)
{
	KzWebKitGtkPrivate *priv = KZ_WEBKIT_GTK_GET_PRIVATE(object);
	if (priv->location)
		g_free(priv->location);
	if (priv->title)
		g_free(priv->title);
	if (priv->link_message)
		g_free(priv->link_message);
	priv->location = NULL;
	priv->title = NULL;
	priv->link_message = NULL;

	if (G_OBJECT_CLASS(kz_webkit_gtk_parent_class)->dispose)
		G_OBJECT_CLASS(kz_webkit_gtk_parent_class)->dispose(object);
}

GtkWidget *
kz_webkit_gtk_new (void)
{
	return GTK_WIDGET(g_object_new(KZ_TYPE_WEBKIT_GTK,
				       "hadjustment", NULL,
				       "vadjustment", NULL,
				       "hscrollbar-policy", GTK_POLICY_AUTOMATIC,
				       "vscrollbar-policy", GTK_POLICY_AUTOMATIC,
				       NULL));
}

static void
load_url (KzEmbed *kzembed, const gchar  *url)
{
	g_return_if_fail(KZ_IS_WEBKIT_GTK(kzembed));

	webkit_web_view_open(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view, url);
}

static void
reload (KzEmbed *kzembed, KzEmbedReloadFlag flags)
{
	g_return_if_fail(KZ_IS_WEBKIT_GTK(kzembed));

	webkit_web_view_reload(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static void
stop_load (KzEmbed *kzembed)
{
	g_return_if_fail(KZ_IS_WEBKIT_GTK(kzembed));

	webkit_web_view_stop_loading(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static void
go_back (KzEmbed *kzembed)
{
	g_return_if_fail(KZ_IS_WEBKIT_GTK(kzembed));

	webkit_web_view_go_back(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static void
go_forward (KzEmbed *kzembed)
{
	g_return_if_fail(KZ_IS_WEBKIT_GTK(kzembed));

	webkit_web_view_go_forward(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static gboolean
can_go_back (KzEmbed *kzembed)
{
	g_return_val_if_fail(KZ_IS_WEBKIT_GTK(kzembed), FALSE);

	return webkit_web_view_can_go_back(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static gboolean
can_go_forward (KzEmbed *kzembed)
{
	g_return_val_if_fail(KZ_IS_WEBKIT_GTK(kzembed), FALSE);

	return webkit_web_view_can_go_forward(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static gboolean
can_cut_selection (KzEmbed *kzembed)
{
	g_return_val_if_fail(KZ_IS_WEBKIT_GTK(kzembed), FALSE);

	return webkit_web_view_can_cut_clipboard(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static gboolean
can_copy_selection (KzEmbed *kzembed)
{
	g_return_val_if_fail(KZ_IS_WEBKIT_GTK(kzembed), FALSE);

	return webkit_web_view_can_copy_clipboard(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static gboolean
can_paste (KzEmbed *kzembed)
{
	g_return_val_if_fail(KZ_IS_WEBKIT_GTK(kzembed), FALSE);

	return webkit_web_view_can_paste_clipboard(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static void
cut_selection (KzEmbed *kzembed)
{
	g_return_if_fail(KZ_IS_WEBKIT_GTK(kzembed));

	webkit_web_view_cut_clipboard(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static void
copy_selection (KzEmbed *kzembed)
{
	g_return_if_fail(KZ_IS_WEBKIT_GTK(kzembed));

	webkit_web_view_copy_clipboard(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static void
paste (KzEmbed *kzembed)
{
	g_return_if_fail(KZ_IS_WEBKIT_GTK(kzembed));

	webkit_web_view_paste_clipboard(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static void
select_all (KzEmbed *kzembed)
{
	g_return_if_fail(KZ_IS_WEBKIT_GTK(kzembed));

	webkit_web_view_select_all(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static const gchar *
get_title (KzEmbed *kzembed)
{
	g_return_val_if_fail(KZ_IS_WEBKIT_GTK(kzembed), NULL);

	return KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->title;
}

static const gchar *
get_location (KzEmbed *kzembed)
{
	g_return_val_if_fail(KZ_IS_WEBKIT_GTK(kzembed), NULL);

	return KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->location;
}

static gchar *
get_link_message (KzEmbed *kzembed)
{
	g_return_val_if_fail(KZ_IS_WEBKIT_GTK(kzembed), NULL);

	return g_strdup(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->link_message);
}

static gdouble
get_progress (KzEmbed *kzembed)
{
	return (gdouble) KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->progress / 100;
}

static gboolean
selection_is_collapsed (KzEmbed *kzembed)
{
	return webkit_web_view_has_selection(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
}

static gboolean
is_loading (KzEmbed *kzembed)
{
	return KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->is_loading;
}

static void
zoom_set (KzEmbed *kzembed, int zoom, gboolean reflow)
{
        webkit_web_view_set_zoom_level(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view,
                                       (gfloat) zoom / 100);
}

static int
zoom_get (KzEmbed *kzembed)
{
	return (gint)webkit_web_view_get_zoom_level(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view) * 100;
}

static gboolean
get_lock (KzEmbed *kzembed)
{
	return KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->lock;
}


static void
set_lock (KzEmbed *kzembed, gboolean lock)
{
	KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->lock = lock;
}

static gboolean
get_allow_javascript (KzEmbed *kzembed)
{
	WebKitWebSettings *settings;
	gboolean allow;

	settings = webkit_web_view_get_settings(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
	if (!settings)
		return FALSE;

	g_object_get(settings, "enable-scripts", &allow, NULL);

	return allow;
}

static void
set_allow_javascript (KzEmbed *kzembed, gboolean allow)
{
	WebKitWebSettings *settings;

	settings = webkit_web_view_get_settings(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
	if (!settings)
		return;

	g_object_set(settings, "enable-scripts", allow, NULL);
}

static gboolean
get_allow_images (KzEmbed *kzembed)
{
	WebKitWebSettings *settings;
	gboolean allow;

	settings = webkit_web_view_get_settings(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
	if (!settings)
		return FALSE;

	g_object_get(settings, "auto-load-images", &allow, NULL);

	return allow;
}

static void
set_allow_images (KzEmbed *kzembed, gboolean allow)
{
	WebKitWebSettings *settings;

	settings = webkit_web_view_get_settings(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
	if (!settings)
		return;

	g_object_set(settings, "auto-load-images", allow, NULL);
}

static gboolean
can_go_nav_link (KzEmbed *kzembed, KzEmbedNavLink link)
{
	return FALSE;
}

static KzNavi *
get_nth_nav_link (KzEmbed *kzembed, KzEmbedNavLink link,
		  guint n)
{
	return NULL;
}

static GList *
get_nav_links (KzEmbed *kzembed, KzEmbedNavLink link)
{
	return NULL;
}

static void
merge_item_to_history (KzBookmarkFolder *history, GList **current_bookmark, WebKitWebHistoryItem *item)
{
	const gchar *title, *uri;
	KzBookmark *bookmark;

	title = webkit_web_history_item_get_title(item);
	uri = webkit_web_history_item_get_uri(item);

	if (!*current_bookmark)
	{
		bookmark = kz_bookmark_new_with_attrs(title, uri, NULL);
		kz_bookmark_folder_append(history, bookmark);
		g_object_unref(bookmark);
	}
	else
	{
		bookmark = KZ_BOOKMARK((*current_bookmark)->data);
		kz_bookmark_set_title(bookmark, title);
		kz_bookmark_set_link(bookmark, uri);
		kz_bookmark_set_last_visited(bookmark, 0);
		*current_bookmark = g_list_next(*current_bookmark);
	}
}

static void
merge_history_items_to_history (KzBookmarkFolder *history, GList **current_bookmark, GList *items)
{
	GList *node;
	for (node = items; node; node = g_list_next(node))
	{
		WebKitWebHistoryItem *item = WEBKIT_WEB_HISTORY_ITEM(node->data);
                if (!item)
                        continue;
		merge_item_to_history(history, current_bookmark, item);
	}
}

static WebKitWebHistoryItem *
ensure_current_item (KzEmbed *kzembed)
{
	WebKitWebBackForwardList *list;
	WebKitWebHistoryItem *current_item = NULL;
	KzWebKitGtkPrivate *priv = KZ_WEBKIT_GTK_GET_PRIVATE(kzembed);

	list = webkit_web_view_get_back_forward_list(priv->web_view);
	if (list)
		current_item = webkit_web_back_forward_list_get_current_item(list);

	if (!current_item)
	{
		current_item = webkit_web_history_item_new_with_data(priv->location,
								     priv->title);
	}

	return current_item;
}

static void
get_history (KzEmbed *kzembed, KzBookmarkFolder *history)
{
	GList *children, *bookmark_node;
	GList *backward_items, *forward_items;
	WebKitWebBackForwardList *list;
	WebKitWebHistoryItem *current_item;

	list = webkit_web_view_get_back_forward_list(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
	if (!list)
		return;

	backward_items = webkit_web_back_forward_list_get_back_list_with_limit(list, 99);
	forward_items = webkit_web_back_forward_list_get_forward_list_with_limit(list, 99);
	current_item = ensure_current_item(kzembed);

	children = kz_bookmark_folder_get_children(history);
	bookmark_node = children;

	merge_history_items_to_history(history, &bookmark_node, backward_items);
	kz_bookmark_folder_set_current_position(history, g_list_length(backward_items));

	merge_item_to_history(history, &bookmark_node, current_item);

	merge_history_items_to_history(history, &bookmark_node, forward_items);

	if (bookmark_node)
		bookmark_node = g_list_last(bookmark_node);
	while (bookmark_node)
	{
		GList *previous_node;
		KzBookmark *child = KZ_BOOKMARK(bookmark_node->data);
		previous_node = g_list_previous(bookmark_node);
		kz_bookmark_folder_remove(history, child);
		bookmark_node = previous_node;
	}

	g_list_free(backward_items);
	g_list_free(forward_items);
	g_list_free(children);
}

static void
set_history (KzEmbed *kzembed, KzBookmarkFolder *history)
{
	WebKitWebBackForwardList *list;
	WebKitWebHistoryItem *item;
	GList *children, *node;
	gint current_pos, total = 0;

	list = webkit_web_view_get_back_forward_list(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
	if (!list)
		return;

	children = kz_bookmark_folder_get_children(history);
	for (node = children; node; node = g_list_next (node))
	{
		KzBookmark *child = KZ_BOOKMARK(node->data);

		item = webkit_web_history_item_new_with_data(kz_bookmark_get_link(child),
							     kz_bookmark_get_title(child));
		webkit_web_back_forward_list_add_item(list, item);
		total++;
	}
	g_list_free(children);

	current_pos = kz_bookmark_folder_get_current_position(history);
	item  = webkit_web_back_forward_list_get_nth_item(list, current_pos - total + 1);
	webkit_web_view_go_to_back_forward_item(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view, item);
}

static void
copy_page (KzEmbed *src_embed, KzEmbed *dest_embed)
{
	WebKitWebBackForwardList *src_list, *dest_list;
	WebKitWebHistoryItem *current_item = NULL;
	gint length, i, current_position;

	src_list = webkit_web_view_get_back_forward_list(KZ_WEBKIT_GTK_GET_PRIVATE(src_embed)->web_view);
	if (!src_list)
		return;
	dest_list = webkit_web_view_get_back_forward_list(KZ_WEBKIT_GTK_GET_PRIVATE(dest_embed)->web_view);
	if (!dest_list)
		return;

	/* Remove all items in dest list. */
	webkit_web_back_forward_list_set_limit(dest_list, 0);

	length = webkit_web_back_forward_list_get_back_length(src_list);
	current_position = length;
	length += webkit_web_back_forward_list_get_forward_length(src_list);
	length++;

	webkit_web_back_forward_list_set_limit(dest_list, 100);
	for (i = 0; i < length; i++) {
		WebKitWebHistoryItem *item;
		item = webkit_web_back_forward_list_get_nth_item(src_list, i);
		webkit_web_back_forward_list_add_item(dest_list, item);
		if (i == current_position)
			current_item = item;
	}
	if (current_item) {
		webkit_web_view_go_to_back_forward_item(KZ_WEBKIT_GTK_GET_PRIVATE(dest_embed)->web_view,
							current_item);
	}

}

static gboolean
shistory_get_pos (KzEmbed *kzembed, gint *pos, gint *count)
{
	WebKitWebBackForwardList *list;

	list = webkit_web_view_get_back_forward_list(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
	if (!list)
		return FALSE;

	*pos = webkit_web_back_forward_list_get_back_length(list);
	*count = *pos + webkit_web_back_forward_list_get_forward_length(list);

	return TRUE;
}

static void
shistory_get_nth (KzEmbed *kzembed, 
		  gint nth,
		  gboolean is_relative,
		  gchar **uri,
		  gchar **title)
{
	WebKitWebBackForwardList *list;
	WebKitWebHistoryItem *item;
	gint back_length;

	list = webkit_web_view_get_back_forward_list(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view);
	if (!list)
		return;
	back_length = webkit_web_back_forward_list_get_back_length(list);
	if (!is_relative)
		nth -= back_length;

	item = webkit_web_back_forward_list_get_nth_item(list, nth);
	if (!item)
		return;

	*uri = g_strdup(webkit_web_history_item_get_uri(item));
	*title = g_strdup(webkit_web_history_item_get_title(item));
}

static gboolean
find (KzEmbed *kzembed, const char *keyword, gboolean backward)
{
	return webkit_web_view_search_text(KZ_WEBKIT_GTK_GET_PRIVATE(kzembed)->web_view,
					   keyword,
					   FALSE,
					   !backward,
					   TRUE);
}

static void
cb_title_changed (WebKitWebView* web_view,
		  WebKitWebFrame* web_frame,
		  const gchar* title,
		  gpointer data)
{
	KzWebKitGtkPrivate *priv = KZ_WEBKIT_GTK_GET_PRIVATE(data);
	if (priv->title)
		g_free(priv->title);
	priv->title = g_strdup(title);

	g_signal_emit_by_name(data, "kz-title");
}

static void
cb_progress_changed (WebKitWebView* web_view,
		     gint progress,
		     gpointer data)
{
	KzWebKitGtkPrivate *priv = KZ_WEBKIT_GTK_GET_PRIVATE(data);
	priv->progress = progress;

	g_signal_emit_by_name(data, "kz-progress");
}

static void
cb_load_committed (WebKitWebView* web_view,
	           WebKitWebFrame* frame,
		   gpointer data)
{
	KzWebKitGtkPrivate *priv = KZ_WEBKIT_GTK_GET_PRIVATE(data);
	const gchar* location = webkit_web_frame_get_uri(frame);
	if (priv->location)
		g_free(priv->location);
	priv->location = g_strdup(location);
	g_signal_emit_by_name(data, "kz-location");
}

static void
cb_load_started (WebKitWebView* web_view,
	         WebKitWebFrame* frame,
		 gpointer data)
{
	KZ_WEBKIT_GTK_GET_PRIVATE(data)->is_loading = TRUE;
	g_signal_emit_by_name(data, "kz-net-start");
}

static void
cb_load_finished (WebKitWebView* web_view,
	          WebKitWebFrame* frame,
		  gpointer data)
{
	KZ_WEBKIT_GTK_GET_PRIVATE(data)->is_loading = FALSE;
	g_signal_emit_by_name(data, "kz-net-stop");
}

static void
cb_hover_link (WebKitWebView* web_view,
	       const gchar *title,
	       const gchar *location,
	       gpointer data)
{
	KzWebKitGtkPrivate *priv = KZ_WEBKIT_GTK_GET_PRIVATE(data);

	if (priv->link_message)
		g_free(priv->link_message);
	priv->link_message = g_strdup(location);

	g_signal_emit_by_name(data, "kz-link-message");
}

static gboolean
cb_navigation_policy_decision_requested (WebKitWebView* web_view,
	                                 WebKitWebFrame *frame,
                                         WebKitNetworkRequest *request,
                                         WebKitWebNavigationAction *action,
                                         WebKitWebPolicyDecision *decision,
                                         gpointer data)
{
	WebKitWebNavigationReason reason;
	gint button;

	reason = webkit_web_navigation_action_get_reason(action);
	button = webkit_web_navigation_action_get_button(action);

	if (reason == WEBKIT_WEB_NAVIGATION_REASON_LINK_CLICKED &&
	    button == 2)
	{
		const char *uri;
		KzEmbed *embed = NULL;

		g_signal_emit_by_name(data, "kz-new-window", &embed);
		uri = webkit_network_request_get_uri(request);
		kz_embed_load_url(embed, uri);

		return TRUE;
	}

	return FALSE;
}

static KzEmbedEventMouse *
create_kz_embed_event_mouse (GdkEventButton *event,
                             KzEmbedEventContext context)
{
	KzEmbedEventMouse *kzevent;

	kzevent = (KzEmbedEventMouse *) kz_embed_event_new(KZ_EMBED_EVENT_MOUSE);
	if (event->state & GDK_SHIFT_MASK)
		kzevent->modifier |= KZ_SHIFT_KEY;
	if (event->state & GDK_CONTROL_MASK)
		kzevent->modifier |= KZ_CTRL_KEY;
	if (event->state & GDK_MOD1_MASK)
		kzevent->modifier |= KZ_ALT_KEY;
	if (event->state & GDK_META_MASK)
		kzevent->modifier |= KZ_META_KEY;
	kzevent->x = event->x;
	kzevent->y = event->y;
        switch (event->button)
        {
        case 1:
                kzevent->button = 1;
                break;
        case 2:
                kzevent->button = 3;
                break;
        case 3:
                kzevent->button = 2;
                break;
        }
	kzevent->cinfo.context = context;

	return kzevent;
}

static gboolean
cb_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	KzEmbedEventMouse *kzevent;
	gint ret = FALSE;

	kzevent = create_kz_embed_event_mouse(event, KZ_CONTEXT_NONE);
	g_signal_emit_by_name(data, "kz-dom-mouse-down", kzevent, &ret);
	kz_embed_event_free((KzEmbedEvent *) kzevent);

	return ret;
}

static gboolean
cb_button_release_event (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	KzEmbedEventMouse *kzevent;
	gint ret = FALSE;

	kzevent = create_kz_embed_event_mouse(event, KZ_CONTEXT_NONE);
	g_signal_emit_by_name(data, "kz-dom-mouse-up", kzevent, &ret);
	kz_embed_event_free((KzEmbedEvent *) kzevent);

	return ret;
}

static void
cb_remove_menu_item (GtkWidget *menu_item, gpointer data)
{
        GtkContainer *container = GTK_CONTAINER(data);
        gtk_container_remove(container, menu_item);
}

static KzEmbedEventContext
get_event_context_from_menu (GtkMenu *menu)
{
        GList *menu_items;
        GtkWidget *first_item;
        GtkWidget *label;
        const gchar *label_text;

        menu_items = gtk_container_get_children(GTK_CONTAINER(menu));
        if (!menu_items)
                return KZ_CONTEXT_NONE;
        first_item = GTK_WIDGET(menu_items->data);
        g_list_free(menu_items);

        label = gtk_bin_get_child(GTK_BIN(first_item));
        label_text = gtk_label_get_text(GTK_LABEL(label));

        if (g_str_equal(label_text, "Open Image in New Window"))
        {
                return KZ_CONTEXT_IMAGE;
        }
        else if (g_str_equal(label_text, "Open Link"))
        {
                return KZ_CONTEXT_LINK;
        }
        return KZ_CONTEXT_NONE;
}

static void
cb_populate_popup (WebKitWebView* web_view,
		   GtkMenu *menu,
		   gpointer data)
{
	KzEmbedEventMouse *kzevent;
        KzEmbedEventContext context;
        GdkEventButton *gdk_event;
        gboolean ret;
        gint x, y;

        context = get_event_context_from_menu(menu);

        gtk_container_forall(GTK_CONTAINER(menu),
                             cb_remove_menu_item, menu);

        gdk_event = ((GdkEventButton*)gdk_event_new(GDK_BUTTON_RELEASE));
        gtk_widget_get_pointer(GTK_WIDGET(web_view),
                               &x, &y);
        gdk_event->x = (gdouble)x;
        gdk_event->y = (gdouble)y;
	gdk_event->button = 3;
	kzevent = create_kz_embed_event_mouse(gdk_event, context);
        if (context & KZ_CONTEXT_LINK)
        {
                KzWebKitGtkPrivate *priv = KZ_WEBKIT_GTK_GET_PRIVATE(data);
                kzevent->cinfo.link = g_strdup(priv->link_message);
        }

	g_signal_emit_by_name(data, "kz-dom-mouse-down", kzevent, &ret);
	g_signal_emit_by_name(data, "kz-dom-mouse-up", kzevent, &ret);
	kz_embed_event_free((KzEmbedEvent *) kzevent);
        gdk_event_free((GdkEvent*)gdk_event);
}

static void
set_font_preferences (KzProfile *profile, WebKitWebSettings *settings)
{
	/* font settings */
	GList *list, *node;
	gint size_variable = -1, size_fixed = -1;
	gint min_size_variable = -1, min_size_fixed = -1;
	gchar langs[1024], *lang = NULL;
	gchar *serif_font, *sans_font, *mono_font;
	double dpi_pixel_ratio = 72.0 / 96.0;

	/* workaround */
	if (kz_profile_get_value(profile, "Language", "accept_languages", 
				&langs, G_N_ELEMENTS(langs),
				KZ_PROFILE_VALUE_TYPE_STRING))
	{
		gchar **split_str;
		split_str = g_strsplit(langs, ",", 1);

		if (split_str[0])
		{
			lang = g_strdup(split_str[0]);
			g_strfreev(split_str);
		}
	}
	if (!lang)
		lang = g_strdup("");

	serif_font = g_strdup_printf("name_serif_%s", lang);
	sans_font = g_strdup_printf("name_sans-serif_%s", lang);
	mono_font = g_strdup_printf("name_monospace_%s", lang);

	list = kz_profile_enum_key(profile, "Font", TRUE);
	for (node = list; node; node = g_list_next(node))
	{
		const gchar *key = (const gchar*)node->data;

		if (!key || !*key) continue;
		if (g_str_has_prefix(key, "size_variable_"))
		{
			gint value;
			kz_profile_get_value(profile, "Font", key,
					&value, sizeof(value),
					KZ_PROFILE_VALUE_TYPE_INT);
			if (value > size_variable)
				size_variable = value;
		}
		else if (g_str_has_prefix(key, "size_fixed_"))
		{
			gint value;
			kz_profile_get_value(profile, "Font", key,
					&value, sizeof(value),
					KZ_PROFILE_VALUE_TYPE_INT);
			if (value > size_fixed)
				size_fixed = value;
		}
		else if (g_str_has_prefix(key, "min-size_variable_"))
		{
			gint value;
			kz_profile_get_value(profile, "Font", key,
					&value, sizeof(value),
					KZ_PROFILE_VALUE_TYPE_INT);
			if (value > min_size_variable)
				min_size_variable = value;
		}
		else if (g_str_has_prefix(key, "min-size_fixed_"))
		{
			gint value;
			kz_profile_get_value(profile, "Font", key,
					&value, sizeof(value),
					KZ_PROFILE_VALUE_TYPE_INT);
			if (value > min_size_fixed)
				min_size_fixed = value;
		}
		else if (!strcmp(key, serif_font))
		{
			gchar value[1024];
			kz_profile_get_value(profile, "Font", key,
					&value, strlen(value)+1,
					KZ_PROFILE_VALUE_TYPE_STRING);
			g_object_set(settings,
				     "serif-font-family", value,
				     "default-font-family", value,
				     NULL);
		}
		else if (!strcmp(key, sans_font))
		{
			gchar value[1024];
			kz_profile_get_value(profile, "Font", key,
					&value, strlen(value)+1,
					KZ_PROFILE_VALUE_TYPE_STRING);
			g_object_set(settings,
				     "sans-serif-font-family", value,
				     NULL);
		}
		else if (!strcmp(key, mono_font))
		{
			gchar value[1024];
			kz_profile_get_value(profile, "Font", key,
					&value, strlen(value)+1,
					KZ_PROFILE_VALUE_TYPE_STRING);
			g_object_set(settings,
				     "monospace-font-family", value,
				     NULL);
		}
	}
	g_list_free(list);
	g_free(serif_font);
	g_free(sans_font);
	g_free(mono_font);

	if (size_variable >= 0)
		g_object_set(settings, "default-font-size", (gint)(size_variable * dpi_pixel_ratio), NULL);
	if (size_fixed >= 0)
		g_object_set(settings, "default-monospace-font-size", (gint)(size_fixed * dpi_pixel_ratio), NULL);
	if (min_size_variable >= 0)
		g_object_set(settings, "minimum-font-size", (gint)(min_size_variable * dpi_pixel_ratio), NULL);
	if (min_size_fixed >= 0)
		g_object_set(settings, "minimum-logical-font-size", (gint)(min_size_fixed * dpi_pixel_ratio), NULL);
}

static void
set_default_preferences (KzWebKitGtk *webkit)
{
	KzProfile *profile = KZ_GET_GLOBAL_PROFILE;
	gchar value[1024];

	WebKitWebSettings *settings;
	settings = webkit_web_view_get_settings(KZ_WEBKIT_GTK_GET_PRIVATE(webkit)->web_view);
	if (!settings)
		return;

	if (kz_profile_get_value(profile, "Language", "charset_default", 
				&value, G_N_ELEMENTS(value),
				KZ_PROFILE_VALUE_TYPE_STRING))
	{
		g_object_set(settings, "default-encoding", value, NULL);
	}

	set_font_preferences (profile, settings);
	
	g_object_set(settings,
		     "enable-plugins", TRUE,
		     "enable-scripts", TRUE,
		     "auto-load-images", TRUE,
		     NULL);
}

